import argparse
import logging

from mcgen.logging import setup_logging
from mcgen.run import run

DEFAULT_JARPATH = "temp/jars/minecraft_server.{version}.jar"
DEFAULT_RAWPATH = "temp/raw/{version}"
DEFAULT_OUTPATH = "temp/out/{version}"
DEFAULT_VERSION = "snapshot"
DEFAULT_MANIFEST = "https://launchermeta.mojang.com/mc/game/version_manifest.json"
DEFAULT_JAVA_CMD = (
    "java -DbundlerMainClass=net.minecraft.data.Main -jar {jar_path} --server --reports"
)
DEFAULT_PROCESSORS = (
    "mcgen.processors.write_version_file",
    "mcgen.processors.convert_json_files",
    "mcgen.processors.simplify_blocks",
    "mcgen.processors.split_registries",
    "mcgen.processors.summarize_data",
    "mcgen.processors.summarize_worldgen",
    "mcgen.processors.create_all_tags_data_pack",
)

ARG_PARSER = argparse.ArgumentParser(
    prog="mcgen",
    description="Download the Minecraft server jar for the specified version,"
    + " invoke the data generator, and process the output.",
)
ARG_PARSER.add_argument(
    "--jarpath",
    default=DEFAULT_JARPATH,
    help=f"Where to download and store the server jar. Default: {DEFAULT_JARPATH}",
)
ARG_PARSER.add_argument(
    "--rawpath",
    default=DEFAULT_RAWPATH,
    help=f"Where to store the raw server-generated files. Default: {DEFAULT_RAWPATH}",
)
ARG_PARSER.add_argument(
    "--outpath",
    default=DEFAULT_OUTPATH,
    help=f"Where to write the final processed output. Default: {DEFAULT_OUTPATH}",
)
ARG_PARSER.add_argument(
    "--version",
    default=DEFAULT_VERSION,
    help="The server version to download and process. Defaults to latest snapshot.",
)
ARG_PARSER.add_argument(
    "--cmd",
    default=DEFAULT_JAVA_CMD,
    help="The java command to invoke the data generator. This string will be formatted"
    + f" with the `jar_path` argument available. Default: {DEFAULT_JAVA_CMD}.",
)
ARG_PARSER.add_argument(
    "--manifest",
    default=DEFAULT_MANIFEST,
    help="Where to fetch the version manifest from. Defaults to Mojang's online copy.",
)
ARG_PARSER.add_argument(
    "--processors",
    default=DEFAULT_PROCESSORS,
    nargs="*",
    help="Which processors to use in processing the raw server-generated files."
    + " Defaults to a set of built-in processors.",
)
ARG_PARSER.add_argument(
    "--log",
    default=logging.WARNING,
    help="The level of verbosity at which to print log messages.",
)
ARGS = ARG_PARSER.parse_args()

setup_logging(ARGS.log)


run(
    jarpath=str(ARGS.jarpath),
    rawpath=str(ARGS.rawpath),
    outpath=str(ARGS.outpath),
    version=str(ARGS.version),
    java_cmd=str(ARGS.cmd),
    manifest_location=str(ARGS.manifest),
    processors=list(ARGS.processors),
)



import json
import logging
from pathlib import Path

from mcgen.context import Context

LOG = logging.getLogger(__name__)


def process(ctx: Context, **options):
    """Create a simplified version of `blocks.json`."""

    blocks_location = "reports/blocks.json"
    LOG.info(f"Simplifying blocks at: {blocks_location}")

    # load blocks.json
    blocks_inpath = ctx.input_dir / blocks_location
    with open(blocks_inpath) as fp:
        blocks_data = json.load(fp)

    # summarize block and block state data
    data = {}
    for block_name, block_data in blocks_data.items():
        data[block_name] = {
            "properties": block_data.get("properties", {}),
            "default": next(
                filter(lambda s: s.get("default", False), block_data["states"])
            ).get("properties", {}),
        }

    # write the simplified blocks files
    node_path = Path("reports/blocks/simplified")
    ctx.write_json_node(data, node_path)
    ctx.write_min_json_node(data, node_path)



import json
import logging
import subprocess
import urllib.request
from importlib import import_module
from pathlib import Path
from typing import Any, List

from mcgen.context import Context

LOG = logging.getLogger(__name__)


def run(
    jarpath: str,
    rawpath: str,
    outpath: str,
    version: str,
    java_cmd: str,
    manifest_location: str,
    processors: List[Any],
):
    LOG.info(f"Using version: {version}")
    LOG.info(f"Using manifest location: {manifest_location}")

    # load the version manifest (remote or local)
    if manifest_location.startswith(("http://", "https://")):
        LOG.info(f"Fetching remote version manifest from: {manifest_location}")
        manifest_raw = urllib.request.urlopen(manifest_location).read()
        manifest = json.loads(manifest_raw)
    else:
        LOG.info(f"Loading local version manifest from: {manifest_location}")
        with open(manifest_location) as manifest_fp:
            manifest = json.load(manifest_fp)

    # resolve the version (release, snapshot, etc)
    if version == "release":
        resolved_version = manifest["latest"]["release"]
        LOG.info(f"Resolved latest release version: {resolved_version}")
    elif version == "snapshot":
        resolved_version = manifest["latest"]["snapshot"]
        LOG.info(f"Resolved latest snapshot version: {resolved_version}")
    else:
        resolved_version = version

    # get the corresponding version entry
    version_entry = next(v for v in manifest["versions"] if v["id"] == resolved_version)
    assert version_entry is not None

    # download the server jar if we don't already have it

    jar_path = Path(jarpath.format(version=resolved_version)).absolute()
    LOG.info(f"Using path for server jar: {jar_path}")

    if jar_path.exists():
        LOG.info(f"Found local server jar at: {jar_path}")

    else:
        LOG.info("Unable to find local server jar; requires download")

        # create parent directories
        jar_path.parent.mkdir(parents=True, exist_ok=True)

        # download the version data
        version_url = version_entry["url"]
        LOG.info(f"Downloading version data from: {version_url}")
        version_data_raw = urllib.request.urlopen(version_url).read()
        version_data = json.loads(version_data_raw)

        # download the server jar
        server_jar_url = version_data["downloads"]["server"]["url"]
        LOG.info(f"Downloading server jar from: {server_jar_url}")
        server_jar_raw = urllib.request.urlopen(server_jar_url).read()
        LOG.info(f"Saving server jar to: {jar_path}")
        with open(jar_path, "wb") as jar_fp:
            jar_fp.write(server_jar_raw)

    # resolve the path to the raw data
    raw_path = Path(rawpath.format(version=resolved_version)).absolute()
    LOG.info(f"Using directory for raw data: {raw_path}")
    if raw_path.exists():
        raise Exception(f"Directory for raw data already exists at: {raw_path}")
    raw_path.mkdir(parents=True, exist_ok=True)

    # collect the raw server-generated data
    LOG.info(f"Storing raw data under: {raw_path}")
    actual_java_cmd = java_cmd.format(jar_path=jar_path)
    LOG.info(f"Invoking server's data generator with: {actual_java_cmd}")
    LOG.info("-" * 80)
    java_proc = subprocess.Popen(actual_java_cmd.split(), cwd=raw_path)
    java_result = java_proc.wait()
    LOG.info("-" * 80)
    LOG.info(f"Server's data generator completed with result: {java_result}")

    # create a folder for the processed output
    out_path = Path(outpath.format(version=resolved_version)).absolute()
    LOG.info(f"Using directory for output: {out_path}")
    if out_path.exists():
        raise Exception(f"Directory for output already exists at: {out_path}")
    out_path.mkdir(parents=True, exist_ok=True)

    # create context and run processors
    LOG.info("Processing data...")
    ctx = Context(
        input_dir=raw_path / "generated",
        output_dir=out_path,
        version=resolved_version,
    )
    for processor in processors:
        try:
            if isinstance(processor, str):
                processor_name = processor
                processor_options = {}
                processor_disabled = processor_name.startswith("!")
            elif isinstance(processor, dict):
                processor_name = processor["name"]
                processor_options = processor.get("options", {})
                processor_disabled = processor.get("disabled", False)
            else:
                raise ValueError(f"Invaid processor: {processor}")
            if processor_disabled:
                continue
            processor_module = import_module(processor_name)
            processor_function = getattr(processor_module, "process")
            LOG.info(f"Running processor: {processor_name}")
            processor_function(ctx, **processor_options)
        except:
            LOG.exception(f"Skipping processor because it caused an error: {processor}")

    LOG.info("Done!")



import json
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List, Tuple

LOG = logging.getLogger(__name__)


@dataclass
class Context:
    input_dir: Path
    output_dir: Path
    version: str
    exclude_dirs: Tuple[str, ...] = (".cache", "tmp")

    def resolve_path(self, relpath: Path) -> Path:
        # assert that the given path is relative
        assert not relpath.is_absolute()
        # convert it to an absolute path
        abspath = (self.output_dir / relpath).resolve()
        # assert that it remains within the output directory
        assert abspath.relative_to(self.output_dir)
        # return the final, resolved path
        return abspath

    def prepare_file_path(self, path: Path) -> Path:
        # resolve the given path
        abspath = self.resolve_path(path)
        # create the containing directory, if it does not already exist
        abspath.parent.mkdir(parents=True, exist_ok=True)
        # return the file path
        return abspath

    def write_json_file(self, data: dict, path: Path):
        file_path = self.prepare_file_path(path)
        LOG.debug(f"Writing JSON file: {file_path}")
        with open(file_path, "w") as fp:
            json.dump(data, fp, indent=2, sort_keys=True)

    def prepare_node_path(self, path: Path, name: str) -> Path:
        # resolve the given path
        abspath = self.resolve_path(path)
        # create the node directory, if it does not already exist
        abspath.mkdir(parents=True, exist_ok=True)
        # return the final file path with the given name
        file_path = abspath / name
        return file_path

    def write_json_node(self, data: dict, path: Path, name="data.json"):
        filepath = self.prepare_node_path(path, name)
        LOG.debug(f"Writing JSON file: {filepath}")
        with open(filepath, "w") as fp:
            json.dump(data, fp, indent=2, sort_keys=True)

    def write_min_json_node(self, data: dict, path: Path, name="data.min.json"):
        filepath = self.prepare_node_path(path, name)
        LOG.debug(f"Writing minified JSON file: {filepath}")
        with open(filepath, "w") as fp:
            json.dump(data, fp, separators=(",", ":"), sort_keys=True)

    def write_values_txt_node(self, values: list, path: Path, name="data.values.txt"):
        filepath = self.prepare_node_path(path, name)
        LOG.debug(f"Writing values TXT file: {filepath}")
        with open(filepath, "w") as fp:
            fp.write("\n".join(values))

    def walk(self, path: Path) -> Iterator[Tuple[str, List[str], List[str]]]:
        for dirpath, dirnames, filenames in os.walk(path):
            dirnames[:] = [d for d in dirnames if d not in self.exclude_dirs]
            yield dirpath, dirnames, filenames



import json
import logging
from pathlib import Path

from mcgen.context import Context

LOG = logging.getLogger(__name__)


def process(ctx: Context, **options):
    """Split `registries.json` into separate files."""

    # load registries.json
    registries_inpath = ctx.input_dir / "reports/registries.json"
    LOG.info(f"Splitting registries at: {registries_inpath}")
    with open(registries_inpath) as fp:
        registries_data = json.load(fp)

    # split each registry into its own set of files
    registries_path = Path("reports/registries")
    for reg_name, registry in registries_data.items():
        reg_entries = registry["entries"]
        LOG.debug(f"Found {len(reg_entries)} entries for registry: {reg_name}")
        reg_shortname = reg_name.split(":")[1]
        reg_node_path = registries_path / reg_shortname
        values = sorted(list(reg_entries.keys()))
        data = {"values": values}
        ctx.write_values_txt_node(values, reg_node_path)
        ctx.write_json_node(data, reg_node_path)
        ctx.write_min_json_node(data, reg_node_path)



import logging
from pathlib import Path

from mcgen.context import Context
from mcgen.utils import summarize_registry

LOG = logging.getLogger(__name__)


def process(ctx: Context, **options):
    """Create a summary of worldgen reports."""

    worldgen_path = Path("reports/worldgen/minecraft")
    LOG.info(f"Summarizing worldgen at: {worldgen_path}")

    # TODO Scan for registries automatically. #enhance

    worldgen_summary = {}

    # dimension
    worldgen_summary["dimension"] = summarize_registry(ctx, worldgen_path / "dimension")

    # dimension_type
    worldgen_summary["dimension_type"] = summarize_registry(
        ctx, worldgen_path / "dimension_type"
    )

    # worldgen/biome
    worldgen_summary["worldgen/biome"] = summarize_registry(
        ctx, worldgen_path / "worldgen/biome"
    )

    # worldgen/configured_carver
    worldgen_summary["worldgen/configured_carver"] = summarize_registry(
        ctx, worldgen_path / "worldgen/configured_carver"
    )

    # worldgen/configured_feature
    worldgen_summary["worldgen/configured_feature"] = summarize_registry(
        ctx, worldgen_path / "worldgen/configured_feature"
    )

    # worldgen/configured_structure_feature
    worldgen_summary["worldgen/configured_structure_feature"] = summarize_registry(
        ctx, worldgen_path / "worldgen/configured_structure_feature"
    )

    # worldgen/noise
    worldgen_summary["worldgen/noise"] = summarize_registry(
        ctx, worldgen_path / "worldgen/noise"
    )

    # worldgen/noise_settings
    worldgen_summary["worldgen/noise_settings"] = summarize_registry(
        ctx, worldgen_path / "worldgen/noise_settings"
    )

    # worldgen/placed_feature
    worldgen_summary["worldgen/placed_feature"] = summarize_registry(
        ctx, worldgen_path / "worldgen/placed_feature"
    )

    # worldgen/processor_list
    worldgen_summary["worldgen/processor_list"] = summarize_registry(
        ctx, worldgen_path / "worldgen/processor_list"
    )

    # worldgen/template_pool
    worldgen_summary["worldgen/template_pool"] = summarize_registry(
        ctx, worldgen_path / "worldgen/template_pool"
    )

    ctx.write_json_node(worldgen_summary, worldgen_path)
    ctx.write_min_json_node(worldgen_summary, worldgen_path)



import logging
from pathlib import Path

from mcgen.context import Context
from mcgen.utils import summarize_registry

LOG = logging.getLogger(__name__)


def process(ctx: Context, **options):
    """Create a summary of each vanilla registry."""

    data_path = Path("data/minecraft")
    LOG.info(f"Summarizing data at: {data_path}")

    # TODO Scan for registries automatically. #enhance

    data_summary = {}

    # advancements
    data_summary["advancements"] = summarize_registry(ctx, data_path / "advancements")

    # loot tables
    data_summary["loot_tables"] = summarize_registry(ctx, data_path / "loot_tables")

    # recipes
    data_summary["recipes"] = summarize_registry(ctx, data_path / "recipes")

    # tags
    data_summary["tags"] = {}
    data_summary["tags"]["blocks"] = summarize_registry(ctx, data_path / "tags/blocks")
    data_summary["tags"]["entity_types"] = summarize_registry(
        ctx, data_path / "tags/entity_types"
    )
    data_summary["tags"]["fluids"] = summarize_registry(ctx, data_path / "tags/fluids")
    data_summary["tags"]["game_events"] = summarize_registry(
        ctx, data_path / "tags/game_events"
    )
    data_summary["tags"]["items"] = summarize_registry(ctx, data_path / "tags/items")

    ctx.write_json_node(data_summary, data_path)
    ctx.write_min_json_node(data_summary, data_path)



import logging
from pathlib import Path
from typing import List

from mcgen.context import Context

LOG = logging.getLogger(__name__)


def summarize_registry(
    ctx: Context,
    registry_path: Path,
    namespace: str = "minecraft",
) -> List[str]:
    LOG.debug(f"Summarizing registry at: {registry_path}")
    registry_indir = ctx.input_dir / registry_path
    unsorted_ids = []
    for dirpath, dirnames, filenames in ctx.walk(registry_indir):
        dirpath_path = Path(dirpath)
        for filename in filenames:
            filepath = dirpath_path / filename
            resource_name = filepath.relative_to(registry_indir).with_suffix("")
            resource_id = f"{namespace}:{resource_name}"
            unsorted_ids.append(resource_id)
    sorted_ids = sorted(unsorted_ids)
    summary = {"values": sorted_ids}
    ctx.write_values_txt_node(sorted_ids, registry_path)
    ctx.write_json_node(summary, registry_path)
    ctx.write_min_json_node(summary, registry_path)
    return sorted_ids



import logging

from mcgen.context import Context

LOG = logging.getLogger(__name__)


def process(
    ctx: Context,
    version_file_location="VERSION.txt",
    **options,
):
    """Write the game version to a file."""
    LOG.info(f"Writing version to: {version_file_location}")
    version_file_path = ctx.output_dir / version_file_location
    with open(version_file_path, "w") as fp:
        fp.write(ctx.version)



import json
import logging
from pathlib import Path

from mcgen.context import Context

LOG = logging.getLogger(__name__)

SUPPORTED_REGISTRIES = {
    "minecraft:block": "blocks",
    "minecraft:entity_type": "entity_types",
    "minecraft:fluid": "fluids",
    "minecraft:game_event": "game_events",
    "minecraft:item": "items",
}


def process(ctx: Context, **options):
    """Create a data pack with tags containing all values of various registries."""

    packs_reldir = Path("datapacks")
    pack_reldir = packs_reldir / "mcdata.all_tags"
    LOG.info(f"Creating 'all tags' data pack at: {pack_reldir}")

    # load registries.json
    registries_inpath = ctx.input_dir / "reports/registries.json"
    with open(registries_inpath) as fp:
        registries_data = json.load(fp)

    # create a new tag for each registry that supports tags
    tags_reldir = pack_reldir / "data/mcdata/tags"
    for reg_name, tag_type in SUPPORTED_REGISTRIES.items():
        registry = registries_data[reg_name]
        reg_entries = registry["entries"]
        tag_file_relpath = tags_reldir / tag_type / "all.json"
        LOG.debug(
            f"Writing tag for registry {reg_name} with {len(reg_entries)} values"
            + " at: {tag_file_name}"
        )
        tag_values = sorted(list(reg_entries.keys()))
        tag_content = {"values": tag_values}
        ctx.write_json_file(tag_content, tag_file_relpath)

    # create a pack.mcmeta
    pack_mcmeta_relpath = pack_reldir / "pack.mcmeta"
    pack_mcmeta_content = {
        "path": {
            # TODO Dynamic pack_format based on extracted jar files? #enhance
            "pack_format": 7,
            "description": "mcdata: all tags",
        }
    }
    ctx.write_json_file(pack_mcmeta_content, pack_mcmeta_relpath)



import json
import logging
from pathlib import Path

from mcgen.context import Context

LOG = logging.getLogger(__name__)


def process(ctx: Context, **options):
    """Convert json files into another form."""

    LOG.info("Converting json files...")

    # convert every file we encounter into multiple forms
    for dirpath, dirnames, filenames in ctx.walk(ctx.input_dir):
        sub_indir = Path(dirpath)
        # determine the corresponding path to the output directory
        sub_reldir = sub_indir.relative_to(ctx.input_dir)
        LOG.debug(f"Converting json files at: {sub_reldir}")
        # process each file
        for filename in filenames:
            if filename.endswith(".json"):
                file_inpath = sub_indir / filename
                with open(file_inpath) as fp:
                    file_data = json.load(fp)
                node_path = (sub_reldir / filename).with_suffix("")
                ctx.write_json_node(file_data, node_path)
                ctx.write_min_json_node(file_data, node_path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mcgen', 'mcgen.processors']

package_data = \
{'': ['*']}

extras_require = \
{'colors': ['colorama>=0.4.3,<0.5.0', 'colorlog>=4.2.1,<5.0.0']}

setup_kwargs = {
    'name': 'mcgen',
    'version': '0.7.0',
    'description': "Python utilities for downloading and processing Minecraft's generated data.",
    'long_description': '# mcgen\n\nPython utilities for downloading and processing Minecraft\'s generated data.\n\n[![PyPI](https://img.shields.io/pypi/v/mcgen.svg)](https://pypi.org/project/mcgen/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/mcgen.svg)](https://pypi.org/project/mcgen/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/arcensoth/mcgen)\n\n## Requirements\n\n- Python 3.8+\n- Java 11+ (for invoking the Minecraft server\'s data generator)\n\n## Installation\n\n```bash\npip install mcgen\n```\n\n## Usage\n\n```bash\npython -m mcgen --help\n```\n\n```\nmcgen [-h] [--jarpath JARPATH] [--rawpath RAWPATH] [--outpath OUTPATH] [--version VERSION] [--manifest MANIFEST] [--processors [PROCESSORS [PROCESSORS ...]]] [--log LOG]\n\nDownload the Minecraft server jar for the specified version, invoke the data generator, and process the output.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --jarpath JARPATH     Where to download and store the server jar. Default: temp/jars/minecraft_server.{version}.jar\n  --rawpath RAWPATH     Where to store the raw server-generated files. Default: temp/raw/{version}\n  --outpath OUTPATH     Where to write the final processed output. Default: temp/out/{version}\n  --version VERSION     The server version to download and process. Defaults to latest snapshot.\n  --manifest MANIFEST   Where to fetch the version manifest from. Defaults to Mojang\'s online copy.\n  --processors [PROCESSORS [PROCESSORS ...]]\n                        Which processors to use in processing the raw server-generated files. Defaults to a set of built-in processors.\n  --log LOG             The level of verbosity at which to print log messages.\n```\n\n## Processors\n\nProcessors are used to process the raw server-generated data and produce output. They are invoked one after the other, in the order they are defined.\n\nTo provide a custom set of processors, use the `--processors` option like so:\n\n```bash\npython -m mcgen --processors mcgen.processors.split_registries mcgen.processors.summarize_data\n```\n\n### Built-in processors\n\nSeveral built-in processors are provided in [`mcgen.processors`](./mcgen/processors):\n\n- [`write_version_file`](./mcgen/processors/write_version_file.py) - Write the game version to a file.\n- [`convert_json_files`](./mcgen/processors/convert_json_files.py) - Convert json files into another form.\n- [`simplify_blocks`](./mcgen/processors/simplify_blocks.py) - Create an optimized summary of blocks.\n- [`split_registries`](./mcgen/processors/split_registries.py) - Split `registries.json` into separate files.\n- [`summarize_data`](./mcgen/processors/summarize_data.py) - Create a summary of each vanilla registry.\n- [`summarize_worldgen`](./mcgen/processors/summarize_worldgen.py) - Create a summary of worldgen reports.\n- [`create_all_tags_data_pack`](./mcgen/processors/create_all_tags_data_pack.py) - Generate a data pack with "all" tags.\n\n## Custom processors\n\nProcessors are Python modules containing a function with the following signature:\n\n```python\ndef process(ctx: Context, **options):\n    ...\n```\n\n- `ctx` contains information about the processing context\n- `options` is a key-value mapping of arbitrary data\n\n## 21w39a\n\nNote that in snapshot 21w39a the `java` command used to invoke the server generator changed:\n\n```bash\n# Prior to 21w39a:\njava -cp {jar_path} net.minecraft.data.Main --server --reports\n\n# From 21w39a onward:\njava -DbundlerMainClass=net.minecraft.data.Main -jar {jar_path} --server --reports\n```\n\nIf you need to generate data for versions prior to 21w39a, use the `--cmd` option with the older `java` command.\n',
    'author': 'Arcensoth',
    'author_email': 'arcensoth@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Arcensoth/mcgen',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging.config

LOG_FORMAT_DETAILED = "%(log_color)s%(asctime)s %(levelname)-8s [%(name)s] %(message)s"

LOG_COLORS = {
    "DEBUG": "cyan",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "red,bg_white",
}


def setup_logging(
    level: str = logging.WARNING,
):
    log_handler = logging.StreamHandler()

    try:
        import colorlog

        log_handler.setFormatter(
            colorlog.ColoredFormatter(fmt=LOG_FORMAT_DETAILED, log_colors=LOG_COLORS)
        )

    except:
        pass

    logging.basicConfig(level=level, handlers=[log_handler])
